import numpy as np
try: import simplejson as json
except ImportError: import json

def process_file(fileName, mask = None, **loadtxt_kwargs):
    """Load Keysight PNA file data into dict.

    Parameters
    ----------
    fileName : string
        Path to data. Actual name of the file should be in the format:
        ``name + '_' + pwr + '_DBM_TEMP_' + temp + '.S2P'`` where:

        - name: Must be 'RES-N' where N is any single character you like.

        - pwr: The power the resonator sees in dBm, three characters Example: '-25'

        - temp: The temperature of the resonator in K, five characters Example: '0.150'

    mask : slice (optional)
        You can pass a slice to cut out some region of data. Example 1: mask = slice(10,-10) will cut out the first 10 and last 10 datapoints. Example 2: maks = slice(None,None,10) will resample your data and take every 10th point. Example 3: mask = slice(None, -50) will cut the last 50 points off your data.

    loadtxt_kwargs : dict (optional)
        This is a pass-through to numpy.loadtxt

    Returns
    -------
    dataDict : dict or ``None``
        Dictionary contains the following keys: 'name', 'temp', 'pwr', 'freq',
        'I', 'Q'. If fileName does not exist, then returns ``None``.

    Note
    ----
    This assumes the data in the file is in three columns in the order frequency, I, Q.

    This is also a terribly written function, and you really should write your own!
    """
    #Find the temperature, power, and name locations from the filename
    tempLoc = fileName.find('TEMP') + 5
    pwrLoc = fileName.find('DBM') - 4
    resNameLoc = fileName.find('RES-')

    if mask is None:
        mask = slice(None, None)
    else:
        assert type(mask) == slice, "mask must be of type slice."

    #Read the temp, pwr, and resName from the filename
    if(fileName[tempLoc + 1] == '.'):
        temp = np.float(fileName[tempLoc:tempLoc+5])

        if fileName[pwrLoc] == '_':
            pwr = np.float(fileName[pwrLoc+1:pwrLoc+3])
        else:
            pwr = np.float(fileName[pwrLoc:pwrLoc+3])

        resName = fileName[resNameLoc:resNameLoc+5]

        #Grab frequency, I, and Q
        fileData = np.loadtxt(fileName, **loadtxt_kwargs)
        freqData = fileData[:,0][mask]
        IData = fileData[:,1][mask]
        QData = fileData[:,2][mask]

        dataDict = {'name':resName,'temp':temp,'pwr':pwr,'freq':freqData,'I':IData,'Q':QData}
        return dataDict
    else:
        
        assert False, "Bad file? " + fileName


import warnings
import matplotlib.pyplot as plt
import matplotlib.gridspec as gs
import matplotlib as mpl
import mpl_toolkits.mplot3d
import numpy as np
import scipy.signal as sps
from .resonator import indexResList

Axes3D = mpl_toolkits.mplot3d.Axes3D

def plotResListData(resList, plot_types=['IQ'], **kwargs):
    r"""Plot resonator data and fits.

    Parameters
    ----------
    resList : list-like
        A list of ``pyres.Resonator`` objects. A single ``Resonator`` object can
        be passed, as long as it is in a list.

    plot_types : list, optional
        A list of plots to create, each one specified by a string. Possible plot
        types are:

        - 'IQ': Plots the real part of the transmission (`I`) vs the imaginary
            part (`Q`). This is the default plot.

        - 'rIQ': Plots the residual of `I` vs the residual of `Q`. This plot is
            only available if the ``do_lmfit`` method of each ``Resonator``
            object has been called. The `I` and `Q` residuals are normalized by
            the uncertainty of the `I` and `Q` data respectively. If this is not
            explicitly supplied, it is calculated by taking the standard
            deviation of the first 10 data points.

        - 'LinMag': Plots the magnitude of the tranmission in Volts vs
            frequency.

        - 'LogMag': Plots the magnitude of the transmission in dB vs frequency.
            ``LogMag = 20*np.log10(LinMag)``.

        - 'rMag': Plots the difference of `LinMag` and the best-fit magnitude vs
            frequency. This plot is only available if the ``do_lmfit`` method of
            each ``Resonator`` object has been called.

        - 'Phase': Plots the phase of the transmision vs frequency.
            ``Phase = np.arctan2(Q, I)``.

        - 'rPhase': Plots the difference of `Phase` and the best-fit phase vs
            frequency. This plot is only available if the ``do_lmfit`` method of
            each ``Resonator`` object has been called.

        - 'uPhase': Plots the unwrapped phase vs frequency.
            ``uPhase = np.unwrap(Phase)``.

        - 'ruPhase': Plots the difference of `uPhase` and the unwrapped best-fit
            phase vs frequency. This plot is only available if the ``do_lmfit``
            method of each ``Resonator`` object has been called.

        - 'I': Plots the real part of the transmission vs frequency.

        - 'rI': Plots the residual of `I` vs frequency. The residual is weighted
            by the uncertainty in `I`. This plot is only available if the
            ``do_lmfit`` method of each ``Resonator`` object has been called.

        - 'Q': Plots the imaginary part of the transmission vs frequency.

        - 'rQ': Plots the residual of `Q` vs frequency. The residual is weighted
            by the uncertainty in `Q`. This plot is only available if the
            ``do_lmfit`` method of each ``Resonator`` object has been called.

    Keyword Arguments
    -----------------

    plot_fits : list-like, optional
        A list of boolean flags, one for each plot type specified. Determines
        whether or not to overplot the best fit on the data. This is only
        effective if the ``do_lmfit`` method of each ``Resonator`` object has
        been called. Default is all False.

    powers : list-like, optional
        A list of power values to plot. Default is to plot all of the unique
        powers that exist in the list of ``Resonator`` objects.

    max_pwr : float
        An upper bound on the powers to plot. Applies to values passed to
        powers also.

    min_pwr : float
        A lower bound on the powers to plot. Applies to values passed to
        powers also.

    temps : list-like, optional
        A list of temperature values to plot. Default is to plot all of the
        unique temperatures that exist in the list of ``Resonator`` obejcts.

    max_temp : float
        An upper bound on the temperatures to plot. Applies to values passed to
        temps also.

    min_temp : float
        A lower bound on the temperatures to plot. Applies to values passed to
        temps also.

    use_itemps : {False, True}, optional
        If a ``ResonatorSweep`` object has been generated from the resList it
        may have added the ``itemp`` attrubute to each ``ResonatorObject`` in
        the list. Specifying ``use_itemps = True`` will force the plotting
        routine to use those tempeartures.

    freq_units : {'GHz', 'Hz', 'kHz', 'MHz', 'THz'}, optional
        The units for the frequency axis, if it exists. Defaul is 'GHz'.

    detrend_phase : {False, True}, optional
        Whether or not to remove a linear trend from the `Phase` data. A typical
        reason for a steep linear offset in the phase is an uncorrected
        electrical delay due to long transmission lines.

    num_cols : int, optional
        The number of columns to include in the grid of subplots. Default is 1.

    fig_size : int, optional
        The size of an individual subplot in inches. Default is 3.

    force_square : {False, True}, optional
        Whether or not to force each subplot axis to be perfectly square.

    show_colorbar : {True, False}, optional
        Whether or not to add a colorbar to the right edge of the figure. The
        colorbar will correspond to the limits of the colored data. Default is
        True.

    color_by : {'temps', 'pwrs'}, optional
        If multiple temperatures and multiple powers are passed, this selects
        which variable will set the color of the plots. Default is 'temps'.

    color_map : str, optional
        The name of any colormap returned by calling
        ``matplotlib.pyplot.colormaps()`` is a valid option. Default is
        'coolwarm'.

    waterfall : numeric, optional
        The value used to space out LogMag vs frequency plots. Currently only
        available for LogMag. Default is 0.

    plot_kwargs : dict, optional
        A dict of keyword arguments to pass through to the individual plots.
        Attempting to set 'color' will result in an error.

    fit_kwargs : dict, optional
        A dict of keyword arguments to pass through to the fit plots. Default is
        a dashed black line.

    Returns
    -------
    figS : ``matplotlib.pyplot.figure``
        A ``matplotlib.pyplot`` figure object.

    """
    #TODO: Add temperature and power masking that makes more sense, like the ability
    #to set a temperature range, or maybe decimate the temperature data. Also
    #need to add the ability to waterfall the mag and phase plots.
    supported_types = ['IQ', 'rIQ', 'LogMag', 'LinMag', 'rMag',
                        'Phase', 'rPhase', 'uPhase', 'ruPhase',
                        'I', 'rI', 'Q', 'rQ']
    assert all(plt_key in supported_types for plt_key in plot_types), "Unsupported plotType requested!"

    #Get a list of unique temps and powers
    powers = []
    temps = []

    #Should we use itemps?
    use_itemps = kwargs.pop('use_itemps', False)
    if use_itemps:
        assert all(hasattr(res, 'itemps') for res in resList), "Could not locate itemp for at least one resonator!"

    for res in resList:
        powers.append(res.pwr)
        if use_itemps:
            temps.append(res.itemp)
        else:
            temps.append(res.temp)

    powers = np.unique(powers)
    temps = np.unique(temps)

    #Optionally override either list
    powers = np.asarray(kwargs.pop('powers', powers))
    temps = np.asarray(kwargs.pop('temps', temps))

    #Get min/max pwrs/temps
    max_pwr = kwargs.pop('max_pwr', np.max(powers)+1)
    min_pwr = kwargs.pop('min_pwr', np.min(powers)-1)

    max_temp = kwargs.pop('max_temp', np.max(temps)+1)
    min_temp = kwargs.pop('min_temp', 0.0)

    #Enforce bounds
    powers = powers[np.where(np.logical_and(powers < max_pwr, powers > min_pwr))]
    temps = temps[np.where(np.logical_and(temps < max_temp, temps > min_temp))]

    #Should we plot best fits?
    plot_fits = kwargs.pop('plot_fits', [False]*len(plot_types))
    assert len(plot_fits) == len(plot_types), "Must specify a fit bool for each plot type."
    if any(plot_fits):
        assert all(res.hasFit for res in resList), "At least one resonator has not been fit yet."

    #Set the units for the frequency axes
    freq_units = kwargs.pop('freq_units', 'GHz')
    assert freq_units in ['Hz', 'kHz', 'MHz', 'GHz', 'THz'], "Unsupported units request!"

    unitsDict = {'Hz':1,
                'kHz':1e3,
                'MHz':1e6,
                'GHz':1e9,
                'THz':1e12}

    #A spacing value to apply to magnitude vs. frequency plots.
    waterfall = kwargs.pop('waterfall', 0)

    #Remove linear phase variation? Buggy...
    detrend_phase = kwargs.pop('detrend_phase', False)

    #Set some plotting defaults
    num_cols = kwargs.pop('num_cols', 1)
    fig_size = kwargs.pop('fig_size', 3)
    force_square = kwargs.pop('force_square', False)
    show_colorbar = kwargs.pop('show_colorbar', True)


    #Should the temperatures or the powers iterate the colors?
    color_by = kwargs.pop('color_by', 'temps')
    assert color_by in ['temps', 'pwrs'], "color_by must be 'temps' or 'pwrs'."

    #Set the colormap: Default to a nice red/blue thing
    #TODO: Allow for custom colormaps (like from Seaborn, etc)

    color_map = kwargs.pop('color_map', 'coolwarm')
    assert color_map in plt.colormaps(), "Unknown colormap provided"
    color_gen = plt.get_cmap(color_map)

    #Any extra kwargs for plotting
    plot_kwargs = kwargs.pop('plot_kwargs', {})
    fit_kwargs = kwargs.pop('fit_kwargs', {'color' : 'k',
                                            'linestyle' : '--',
                                            'linewidth' : 1.5})

    if kwargs:
        raise NameError("Unknown keyword argument: " + kwargs.keys()[0])


    #Set up the figure
    figS = plt.figure()

    #Calculate rows for figure size
    num_rows = int(np.ceil(1.0*len(plot_types)/num_cols))

    #Set figure size, including some extra spacing for the colorbar
    #0.1 is the extra space for the colorbar.
    #*1.2 is the extra padding for the axis labels
    #15:1 is the ratio of axis width for regular axes to colorbar axis
    if show_colorbar:
        figS.set_size_inches(fig_size*(num_cols+0.1)*1.2, fig_size*num_rows)

        #Initialize the grid for plotting
        plt_grid = gs.GridSpec(num_rows, num_cols+1, width_ratios=[15]*num_cols+[1])
    else:
        figS.set_size_inches(fig_size*(num_cols)*1.2, fig_size*num_rows)

        #Initialize the grid for plotting
        plt_grid = gs.GridSpec(num_rows, num_cols)

    #Initialize a dictionary of axes corresponding to plot types
    axDict = {}

    #Set up axes and make labels
    for ix, key in enumerate(plot_types):

        iRow = ix//num_cols
        iCol = ix%num_cols

        ax = figS.add_subplot(plt_grid[iRow, iCol])

        if key == 'IQ':
            ax.set_xlabel('I (Volts)')
            ax.set_ylabel('Q (Volts)')

        if key == 'rIQ':
            ax.set_xlabel('Residual of I / $\sigma_\mathrm{I}$')
            ax.set_ylabel('Residual of Q / $\sigma_\mathrm{Q}$')

        if key in ['LogMag', 'LinMag', 'rMag', 'Phase', 'rPhase', 'uPhase',
                    'ruPhase', 'I', 'Q', 'rQ', 'rI']:
            ax.set_xlabel('Frequency (' + freq_units + ')')

        if key == 'LinMag':
            ax.set_ylabel('Magnitude (Volts)')

        if key == 'LogMag':
            ax.set_ylabel('Magnitude (dB)')

        if key == 'rMag':
            ax.set_ylabel('Residual of Magnitude (Volts)')

        if key == 'Phase':
            ax.set_ylabel('Phase (Radians)')

        if key == 'rPhase':
            ax.set_ylabel('Residual of Phase (Radians)')

        if key == 'uPhase':
            ax.set_ylabel('Unwrapped Phase (Radians)')

        if key == 'ruPhase':
            ax.set_ylabel('Residual of unwrapped Phase (Radians)')

        if key == 'I':
            ax.set_ylabel('I (Volts)')

        if key == 'Q':
            ax.set_ylabel('Q (Volts)')

        if key == 'rI':
            ax.set_ylabel('Residual of I / $\sigma_\mathrm{I}$')

        if key == 'rQ':
            ax.set_ylabel('Residual of Q / $\sigma_\mathrm{Q}$')

        #Stuff the axis into the axis dictionary
        axDict[key] = ax

    #Plot the data
    for pwr in powers:
        #The waterfall index should be reset each iteration
        wix = 0
        for temp in temps:
            #Grab the right resonator from the list
            resIndex = indexResList(resList, temp, pwr, itemp=use_itemps)

            #Color magic!
            if color_by == 'temps':
                if len(temps) > 1:
                    plt_color = color_gen(temp*1.0/max(temps))
                else:
                    plt_color = color_gen(0)
            elif color_by == 'pwrs':
                if len(powers) > 1:
                    plt_color = color_gen(1-((max(powers)-pwr)*1.0/(max(powers)-min(powers))))
                else:
                    plt_color = color_gen(0)

            #Not every temp/pwr combo corresponds to a resonator. Ignore missing ones.
            if resIndex is not None:
                res = resList[resIndex]
                scaled_freq = res.freq/unitsDict[freq_units]

                for key, ax in axDict.items():
                    pix = plot_types.index(key)
                    plot_fit = plot_fits[pix]
                    if key == 'IQ':
                        ax.plot(res.I, res.Q, color=plt_color, **plot_kwargs)
                        if plot_fit:
                            ax.plot(res.resultI, res.resultQ, **fit_kwargs)

                    if key == 'rIQ':
                        ax.plot(res.residualI, res.residualQ, color=plt_color, **plot_kwargs)

                    if key == 'LogMag':
                        ax.plot(scaled_freq, res.logmag+wix*waterfall, color=plt_color, **plot_kwargs)
                        if plot_fit:
                            ax.plot(scaled_freq, 20*np.log10(res.resultMag)+wix*waterfall, **fit_kwargs)
                        #Step the waterfall plot
                        wix+=1

                    if key == 'LinMag':
                        ax.plot(scaled_freq, res.mag, color=plt_color, **plot_kwargs)
                        if plot_fit:
                            ax.plot(scaled_freq, res.resultMag, **fit_kwargs)

                    if key == 'rMag':
                        ax.plot(scaled_freq, res.resultMag-res.mag, color=plt_color, **plot_kwargs)

                    if key == 'Phase':
                        if detrend_phase:
                            ax.plot(scaled_freq, sps.detrend(res.phase), color=plt_color, **plot_kwargs)
                            if plot_fit:
                                ax.plot(scaled_freq, sps.detrend(res.resultPhase), **fit_kwargs)
                        else:
                            ax.plot(scaled_freq, res.phase, color=plt_color, **plot_kwargs)
                            if plot_fit:
                                ax.plot(scaled_freq, res.resultPhase, **fit_kwargs)

                    if key == 'rPhase':
                        ax.plot(scaled_freq, res.resultPhase-res.phase, color=plt_color, **plot_kwargs)

                    if key == 'uPhase':
                        if detrend_phase:
                            ax.plot(scaled_freq, sps.detrend(res.uphase), color=plt_color, **plot_kwargs)
                            if plot_fit:
                                ax.plot(scaled_freq, sps.detrend(np.unwrap(res.resultPhase)), **fit_kwargs)
                        else:
                            ax.plot(scaled_freq, res.uphase, color=plt_color, **plot_kwargs)
                            if plot_fit:
                                ax.plot(scaled_freq, np.unwrap(res.resultPhase), **fit_kwargs)

                    if key == 'ruPhase':
                        ax.plot(scaled_freq, np.unwrap(res.resultPhase)-res.uphase, color=plt_color, **plot_kwargs)

                    if key == 'I':
                        ax.plot(scaled_freq, res.I, color=plt_color, **plot_kwargs)
                        if plot_fit:
                            ax.plot(scaled_freq, res.resultI, **fit_kwargs)

                    if key == 'rI':
                        ax.plot(scaled_freq, res.residualI, color=plt_color, **plot_kwargs)

                    if key == 'Q':
                        ax.plot(scaled_freq, res.Q, color=plt_color, **plot_kwargs)
                        if plot_fit:
                            ax.plot(scaled_freq, res.resultQ, **fit_kwargs)

                    if key == 'rQ':
                        ax.plot(scaled_freq, res.residualQ, color=plt_color, **plot_kwargs)

                    xticks = ax.get_xticks()
                    ax.set_xticklabels(xticks,rotation=45)

                    if force_square:
                        #Make the plot a square
                        x1, x2 = ax.get_xlim()
                        y1, y2 = ax.get_ylim()

                        #Explicitly passing a float to avoid a warning in matplotlib
                        #when it gets a numpy.float64
                        ax.set_aspect(float((x2-x1)/(y2-y1)))

    #Add the colorbar
    if show_colorbar:
        #Set limits for the min and max colors
        if color_by == 'temps':
            cbar_norm= mpl.colors.Normalize(vmin=0, vmax=max(temps))
            cbar_units = 'Kelvin'
        elif color_by == 'pwrs':
            cbar_norm = mpl.colors.Normalize(vmin=min(powers), vmax=max(powers))
            cbar_units = 'dB'

        #Make an axis that spans all rows for the colorbar
        cax = figS.add_subplot(plt_grid[:, num_cols])

        #Plot and label the colorbar
        cbar_plot = mpl.colorbar.ColorbarBase(cax, cmap=color_gen, norm=cbar_norm)
        cbar_plot.set_label(cbar_units)

    #Clean up the subfigs and make sure nothing overlaps
    figS.tight_layout()

    return figS

def plotResSweepParamsVsX(resSweep, plot_keys=None, ignore_keys=None, xvals='temperature', **kwargs):
    r"""Plot parameter data vs temperature from a ResonatorSweep object.

    Parameters
    ----------
    resSweep : ``scraps.ResonatorSweep`` object or list of objects
        The object containing the data you want to look at. It is also possible
        to pass a list of these objects and it will combine them into one plot.

    plot_keys : list-like (optional)
        A list of strings corresponding to avaiable plot data. The available
        keys depend on your parameter definitions and may be found by executing
        ``print(resSweep.keys())``. Some keys may point to empty (NaN) objects.
        Default is to plot all of the keys that exist. If you pass plot_keys
        you may not pass ignore_ignore keys.

    ignore_keys : list-like (optional)
        A list of strings corresponding to plots that should not be made. This
        is useful if you want to plot most of the avaialble data, but ignore one
        or two sets of data. Default is ``None``. If you pass ignore_keys you
        may not pass plot_keys.

    xvals : string
        What axis you want to plot by. Deafult is 'temperature'. Could also be
        'power'. Future releases may include other options.

    color_by : string
        Set this to 'index' if you pass multiple res sweeps and want to color them by index.

    Keyword Arguments
    -----------------
    plot_labels : list-like
        A list of strings to use to label the y-axes of the plots. There must be
        one for each plot requested. ``None`` is acceptable for any position in
        the list and will default to using the key as the label. Default is to
        use the key as the label.

    unit_multipliers : list-like
        A list of numbers to multiply against the y-axes data. There must be one
        for each plot requested. ``None`` is acceptable for any position in the
        list and will default to 1. Default is 1.

    fitter : string {'lmfit', 'emcee'}
        Which fit data to use when overlaying best fits. Default is 'lmfit'.

    num_cols : int
        The number of columns to create in the plot grid. Default is 1. The
        number of rows will be calculated based on num_cols and the number of
        requested plots.

    powers : list
        List of powers to plot. Default is to plot all available. If xvals is
        'power' this kwarg is ignored.

    temperatures : list
        List of temperatures to plot. Default is to plot all available. If xvals
        is 'temperature' this kwarg is ignored.

    xmax : numeric
        Don't plot any xvals above this value. Default is infinity.

    xmin : numeric
        Don't plot any xvals below this value. Default is 0 for temperature and
        -infinity for power.

    errorbars: {None, 'lmfit', 'emcee'}
        Add error bars to the data. Pulls from either the least-squares or the
        MCMC fits. Default is None.

    fig_size : numeric
        Size in inches for each plot in the figure.

    color_map : string
        Specifies the colormap to use. Any value in ``matplotlib.pyplot.colormaps()``
        is a valid option.

    show_colorbar : {True, False}, optional
        Whether or not to add a colorbar to the right edge of the figure. The
        colorbar will correspond to the limits of the colored data. Default is
        True.

    force_square : bool
        Whether or not to force each subplot to have perfectly square axes.

    plot_kwargs : dict or list of dicts
        Dict of keyword args to pass through to the plotting function. Default
        is {'linestyle':'--', label='Power X dB'}. If errorbars is not None,
        then default linestyle is {'linestyle':'o'}. Attempting to set 'color'
        or 'yerr' will result in an exception. Use the color_map and errorbars
        keywords to set those. If you passed in a list of objects to resSweep,
        then you can also pass a list of plot_kwargs, one for each sweep object.

    """

    if type(resSweep) is not list:
        resSweep = [resSweep]

    for rS in resSweep:

        #This will really only work for sure if block is sucessful
        assert rS.smartindex == 'block', "index must be 'block' for plotting to work."
        #TODO: fix for other smartindex types

    #set defaults
    plot_labels = kwargs.pop('plot_labels', None)

    unit_multipliers = kwargs.pop('unit_multipliers', None)

    #Which fit data should be plot? lmfit or emcee?
    fitter = kwargs.pop('fitter', 'lmfit')

    #Number of columns
    num_cols = int(kwargs.pop('num_cols', 4))

    #Powers to plot
    powers = list(kwargs.pop('powers', []))

    #Temperatures to plot
    temps = list(kwargs.pop('temps', []))

    if xvals == 'temperature':
        #Just use all the powers that exist period!
        if len(powers) == 0:
            for rS in resSweep:
                powers.extend(rS.pvec)
            powers = list(set(powers))
        else:
            for rS in resSweep:
                assert any(p in rS.pvec for p in powers), "No data exists at any requested power."

    if xvals == 'power':
        #Just use all the temperatures that exist period!
        if len(temps) == 0:
            for rS in resSweep:
                temps.extend(rS.tvec)
            temps = list(set(temps))
        else:
            for rS in resSweep:
                assert any(t in rS.tvec for t in temps), "No data exists at any requested temperature."

    #Set up the x-axis mask
    xmax = kwargs.pop('xmax', None)
    xmin = kwargs.pop('xmin', None)

    if xmin is None:
        if xvals == 'temperature':
            xmin = 0
        elif xvals == 'power':
            xmin = -np.inf

    #Just use the biggest temperature in any set if max isn't passed
    if xmax is None:
        xmax = np.inf

    xMask = []
    for rS in resSweep:
        if xvals == 'temperature':
            xvec = rS.tvec
        elif xvals == 'power':
            xvec = rS.pvec
        xMask.append((xvec >= xmin) * (xvec <= xmax))

    color_by = kwargs.pop('color_by', None)

    if color_by is None:
        if xvals == 'temperature':
            color_by = 'power'
        elif xvals == 'power':
            color_by = 'temperature'

    #Very early errobar code. Still in beta.
    errorbars = kwargs.pop('errorbars', None)
    assert errorbars in [None, 'lmfit', 'emcee'], "Invalid option for errorbars. Try None, 'lmfit', or 'emcee'."

    #Figure out which parameters to plot
    if ignore_keys is None:
        ignore_keys = ['listIndex',
                        'temps']
    else:
        assert plot_keys is None, "Either pass plot_keys or ignore_keys, not both."
        for rS in resSweep:
            assert all(key in rS.keys() for key in ignore_keys), "Unknown key in ignore_keys"
        ignore_keys.append('listIndex')
        ignore_keys.append('temps')

    if plot_keys is None:
        plot_keys = []

        for rS in resSweep:
            plot_keys.extend(set(rS.keys()))
            plot_keys = set(set(plot_keys)-set(ignore_keys))
    else:
        for rS in resSweep:
            assert any(key in rS.keys() for key in plot_keys), "No data corresponding to any plot_key"

    #Some general defaults
    fig_size = kwargs.pop('fig_size', 3)

    force_square = kwargs.pop('force_square', False)

    #Set the colormap: Default to viridis
    color_map = kwargs.pop('color_map', None)
    if color_map is None:
        if color_by == 'index':
            color_map = 'tab10'
        elif xvals == 'temperature':
            color_map = 'viridis'
        elif xvals == 'power':
            color_map = 'coolwarm'

    assert color_map in plt.colormaps(), "Unknown colormap provided"
    color_gen = plt.get_cmap(color_map)

    #Set whether to show the colorbar
    show_colorbar = kwargs.pop('show_colorbar', True)

    #Defaults for this are set later
    plot_kwargs = kwargs.pop('plot_kwargs', [{}]*len(resSweep))

    if type(plot_kwargs) == dict:
        plot_kwargs = [plot_kwargs]*len(resSweep)

    assert type(plot_kwargs) == list, "Must pass list of plot_kwargs of same length as list of resSweeps"

    #Unknown kwargs are discouraged
    if kwargs:
        raise NameError("Unknown keyword argument: " + list(kwargs.keys())[0])

    #Set up the figure
    figS = plt.figure()

    num_keys = len(plot_keys)

    #Don't need more columns than plots
    if num_keys < num_cols:
        num_cols = num_keys

    #Calculate rows for figure size
    num_rows = int(np.ceil(num_keys/num_cols))

    #Calculate rows for figure size
    num_rows = int(np.ceil(1.0*num_keys/num_cols))

    #Set figure size, including some extra spacing for the colorbar
    #0.1 is the extra space for the colorbar.
    #*1.2 is the extra padding for the axis labels
    #15:1 is the ratio of axis width for regular axes to colorbar axis
    if show_colorbar:
        figS.set_size_inches(fig_size*(num_cols+0.1)*1.2, fig_size*num_rows)

        #Initialize the grid for plotting
        plt_grid = gs.GridSpec(num_rows, num_cols+1, width_ratios=[15]*num_cols+[1])
    else:
        figS.set_size_inches(fig_size*(num_cols)*1.2, fig_size*num_rows)

        #Initialize the grid for plotting
        plt_grid = gs.GridSpec(num_rows, num_cols)

    #Loop through all the keys in the ResonatorSweep object and plot them
    for ix, key in enumerate(plot_keys):

        iRow = int(ix/num_cols)
        iCol = ix%num_cols

        axs = figS.add_subplot(plt_grid[iRow, iCol])

        if unit_multipliers is not None:
            mult = unit_multipliers[ix]
        else:
            mult = 1

        if xvals == 'power':
            iterator = temps
        elif xvals == 'temperature':
            iterator = powers

        for itr in iterator:

            if (xvals == 'temperature') and (color_by != 'index'):
                if len(powers) > 1:
                    plt_color = color_gen(1-((max(powers)-itr)*1.0/(max(powers)-min(powers))))
                else:
                    plt_color = color_gen(0)

            elif (xvals == 'power') and (color_by != 'index'):
                if len(temps) > 1:
                    plt_color = color_gen(1-((max(temps)-itr)*1.0/(max(temps)-min(temps))))
                else:
                    plt_color = color_gen(0)

            for rix, rS in enumerate(resSweep):

                if color_by == 'index':
                    plt_color = color_gen(rix)

                if xvals == 'temperature':
                    if itr in rS.pvec:
                        x_data = rS.tvec[xMask[rix]]
                        plt_data = mult*rS[key].loc[xMask[rix], itr].values
                    else:
                        plt_data = None

                elif xvals == 'power':
                    if itr in rS.tvec:
                        x_data = rS.pvec[xMask[rix]]
                        plt_data = mult*rS[key].loc[itr, xMask[rix]].values
                    else:
                        plt_data = None

                if plt_data is not None:

                    if 'label' not in plot_kwargs[rix].keys():
                        if color_by == 'index':
                            plot_kwargs[rix]['label'] = 'Index: '+str(rix)
                        else:
                            plot_kwargs[rix]['label'] = xvals+ ": "+str(itr)

                    if 'linestyle' not in plot_kwargs[rix].keys():
                        if errorbars is not None:
                            plot_kwargs[rix]['marker'] = 'o'
                        else:
                            plot_kwargs[rix]['linestyle'] = '--'

                    if errorbars is None:
                        axs.plot(x_data ,plt_data, color=plt_color, **plot_kwargs[rix])
                    elif errorbars == 'lmfit':
                        #lmfit uncertainty was stored in the _sigma key, so just grab it back out
                        if xvals == 'temperature':
                            plt_err = mult*rS[key + '_sigma'].loc[xMask[rix], itr].values
                        elif xvals == 'power':
                            plt_err = mult*rS[key + '_sigma'].loc[itr, xMask[rix]].values

                        axs.errorbar(x_data, plt_data, yerr=plt_err, color=plt_color, **plot_kwargs[rix])
                    elif errorbars == 'emcee':
                        #emcee uncertainty was placed in the _sigma_plus_mc and _sigma_minus_mc keys
                        if xvals == 'temperature':
                            plt_err_plus = mult*rS[key + '_sigma_plus_mc'].loc[xMask[rix], itr].values
                            plt_err_minus = mult*rS[key + '_sigma_minus_mc'].loc[xMask[rix], itr].values
                        elif xvals == 'power':
                            plt_err_plus = mult*rS[key + '_sigma_plus_mc'].loc[itr, xMask[rix]].values
                            plt_err_minus = mult*rS[key + '_sigma_minus_mc'].loc[itr, xMask[rix]].values

                        plt_err = [plt_err_plus, plt_err_minus]
                        axs.errorbar(x_data, plt_data, yerr=plt_err, color=plt_color, **plot_kwargs[rix])

        if xvals == 'temperature':
            axs.set_xlabel('Temperature (mK)')
        elif xvals == 'power':
            axs.set_xlabel('Power (dB)')

        if plot_labels is not None:
            axs.set_ylabel(plot_labels[ix])
        else:
            axs.set_ylabel(key)
            
        #No idea why this is necessary, but it all falls apart without it
        axs.set_xlim(np.min(x_data), np.max(x_data))
        xticks = axs.get_xticks()
        axs.set_xticks(xticks)
        axs.set_xticklabels(xticks,rotation=45)

        if force_square:
            #Make the plot a square
            x1, x2 = axs.get_xlim()
            y1, y2 = axs.get_ylim()

            #Explicitly passing a float to avoid an warning in matplotlib
            #when it gets a numpy.float64
            axs.set_aspect(float((x2-x1)/(y2-y1)))

        #Stick some legends where they won't crowd too much
        # if key == 'f0' or key == 'fmin':
        #     axs.legend(loc='best')
    if show_colorbar:
        if color_by == 'index':
            cbar_norm = mpl.colors.BoundaryNorm(range(len(resSweep)+1), len(resSweep))
            cbar_units = 'index'
        else:
            cbar_norm = mpl.colors.Normalize(vmin=min(iterator), vmax=max(iterator))

            if xvals == 'temperature':
                cbar_units = 'dB'
            elif xvals == 'power':
                cbar_units = 'mK'

        #Make an axis that spans all rows
        cax = figS.add_subplot(plt_grid[:, num_cols])

        #Plot and label
        cbar_plot = mpl.colorbar.ColorbarBase(cax, cmap=color_gen, norm=cbar_norm)
        cbar_plot.set_label(cbar_units)

    figS.tight_layout()
    return figS



def plotResSweepParamsVsTemp(resSweep, plot_keys=None, ignore_keys=None, **kwargs):
    r"""Plot parameter data vs temperature from a ResonatorSweep object.

    Parameters
    ----------
    resSweep : ``scraps.ResonatorSweep`` object or list of objects
        The object containing the data you want to look at. It is also possible
        to pass a list of these objects and it will combine them into one plot.

    plot_keys : list-like (optional)
        A list of strings corresponding to avaiable plot data. The available
        keys depend on your parameter definitions and may be found by executing
        ``print(resSweep.keys())``. Some keys may point to empty (NaN) objects.
        Default is to plot all of the keys that exist. If you pass plot_keys
        you may not pass ignore_ignore keys.

    ignore_keys : list-like (optional)
        A list of strings corresponding to plots that should not be made. This
        is useful if you want to plot most of the avaialble data, but ignore one
        or two sets of data. Default is ``None``. If you pass ignore_keys you
        may not pass plot_keys.

    Keyword Arguments
    -----------------
    plot_labels : list-like
        A list of strings to use to label the y-axes of the plots. There must be
        one for each plot requested. ``None`` is acceptable for any position in
        the list and will default to using the key as the label. Default is to
        use the key as the label.

    unit_multipliers : list-like
        A list of numbers to multiply against the y-axes data. There must be one
        for each plot requested. ``None`` is acceptable for any position in the
        list and will default to 1. Default is 1.

    fitter : string {'lmfit', 'emcee'}
        Which fit data to use when overlaying best fits. Default is 'lmfit'.

    num_cols : int
        The number of columns to create in the plot grid. Default is 1. The
        number of rows will be calculated based on num_cols and the number of
        requested plots.

    powers : list
        List of powers to plot. Default is to plot all available.

    max_temp : numeric
        Don't plot any temperatures above this value. Default is infinity.

    min_temp : numeric
        Don't plot any temperatures below this value. Default is 0.

    errorbars: {None, 'lmfit', 'emcee'}
        Add error bars to the data. Pulls from either the least-squares or the
        MCMC fits. Default is None.

    fig_size : numeric
        Size in inches for each plot in the figure.

    color_map : string
        Specifies the colormap to use. Any value in ``matplotlib.pyplot.colormaps()``
        is a valid option.

    show_colorbar : {True, False}, optional
        Whether or not to add a colorbar to the right edge of the figure. The
        colorbar will correspond to the limits of the colored data. Default is
        True.

    force_square : bool
        Whether or not to force each subplot to have perfectly square axes.

    plot_kwargs : dict or list of dicts
        Dict of keyword args to pass through to the plotting function. Default
        is {'linestyle':'--', label='Power X dB'}. If errorbars is not None,
        then default linestyle is {'linestyle':'o'}. Attempting to set 'color'
        or 'yerr' will result in an exception. Use the color_map and errorbars
        keywords to set those. If you passed in a list of objects to resSweep,
        then you can also pass a list of plot_kwargs, one for each sweep object.

    """

    warnings.warn("This function has been deprecated in favor of plotResSweepParamsVsX", DeprecationWarning)

    xmin = kwargs.pop('min_temp', None)
    xmax = kwargs.pop('max_temp', None)

    if xmin is not None:
        kwargs['xmin'] = xmin

    if xmax is not None:
        kwargs['xmax'] = xmax

    fig = plotResSweepParamsVsX(resSweep, plot_keys, ignore_keys, xvals='temperature', **kwargs)
    return fig

def plotResSweepParamsVsPwr(resSweep, plot_keys=None, ignore_keys=None, **kwargs):
    r"""Plot parameter data vs power from a ResonatorSweep object.

    Parameters
    ----------
    resSweep : ``scraps.ResonatorSweep`` object or list of objects
        The object containing the data you want to look at. It is also possible
        to pass a list of these objects and it will combine them into one plot.

    plot_keys : list-like (optional)
        A list of strings corresponding to avaiable plot data. The available
        keys depend on your parameter definitions and may be found by executing
        ``print(resSweep.keys())``. Some keys may point to empty (NaN) objects.
        Default is to plot all of the keys that exist. If you pass plot_keys
        you may not pass ignore_ignore keys.

    ignore_keys : list-like (optional)
        A list of strings corresponding to plots that should not be made. This
        is useful if you want to plot most of the avaialble data, but ignore one
        or two sets of data. Default is ``None``. If you pass ignore_keys you
        may not pass plot_keys.

    Keyword Arguments
    -----------------
    plot_labels : list-like
        A list of strings to use to label the y-axes of the plots. There must be
        one for each plot requested. ``None`` is acceptable for any position in
        the list and will default to using the key as the label. Default is to
        use the key as the label.

    unit_multipliers : list-like
        A list of numbers to multiply against the y-axes data. There must be one
        for each plot requested. ``None`` is acceptable for any position in the
        list and will default to 1. Default is 1.

    fitter : string {'lmfit', 'emcee'}
        Which fit data to use when overlaying best fits. Default is 'lmfit'.

    num_cols : int
        The number of columns to create in the plot grid. Default is 1. The
        number of rows will be calculated based on num_cols and the number of
        requested plots.

    temps : list
        List of temperatures to plot. Default is to plot all available.

    max_power : numeric
        Don't plot any temperatures above this value. Default is infinity.

    min_power : numeric
        Don't plot any temperatures below this value. Default is 0.

    errorbars: {None, 'lmfit', 'emcee'}
        Add error bars to the data. Pulls from either the least-squares or the
        MCMC fits. Default is None.

    fig_size : numeric
        Size in inches for each plot in the figure.

    color_map : string
        Specifies the colormap to use. Any value in ``matplotlib.pyplot.colormaps()``
        is a valid option.

    show_colorbar : {True, False}, optional
        Whether or not to add a colorbar to the right edge of the figure. The
        colorbar will correspond to the limits of the colored data. Default is
        True.

    force_square : bool
        Whether or not to force each subplot to have perfectly square axes.

    plot_kwargs : dict or list of dicts
        Dict of keyword args to pass through to the plotting function. Default
        is {'linestyle':'--', label='temperature X mK'}. If errorbars is not None,
        then default linestyle is {'linestyle':'o'}. Attempting to set 'color'
        or 'yerr' will result in an exception. Use the color_map and errorbars
        keywords to set those. If you passed in a list of objects to resSweep,
        then you can also pass a list of plot_kwargs, one for each sweep object.

    """

    warnings.warn("This function has been deprecated in favor of plotResSweepParamsVsX", DeprecationWarning)

    xmin = kwargs.pop('min_power', None)
    xmax = kwargs.pop('max_power', None)

    if xmin is not None:
        kwargs['xmin'] = xmin

    if xmax is not None:
        kwargs['xmax'] = xmax

    fig = plotResSweepParamsVsX(resSweep, plot_keys, ignore_keys, xvals='power', **kwargs)
    return fig



def plotResSweep3D(resSweep, plot_keys, **kwargs):
    r"""Make 3D surface or mesh plots of any key in the ``ResonatorSweep`` object as functions
    of temperature and power.

    Parameters
    ----------
    resSweep : ``ResonatorSweep`` object
        A ``pyres.ResonatorSweep`` object containing all of the data to be
        plotted.

    plot_keys : list-like
        List of strings where each string is a key corresponding to a plot that
        should be made. For a list of acccetable keys, run ``print
        resSweep.keys()``.

    Keyword arguments
    -----------------

    min_temp : numeric (optional)
        The minimum temperature to plot. Defaults to ``min(temperatures)``.

    max_temp : numeric (optional)
        The maximum temperature to plot. Defaults to ``max(temperatres)``.

    min_pwr : numeric (optional)
        The minimum power to plot. Defaults to ``min(powers)``.

    max_pwr : numeric (optional)
        The maximum power to plot. Defauts to ``max(powers)``.

    unit_multipliers : list (optional)
        Values to scale the z-axis by. Default is 1.
        ``len(unit_multipliers) == len(plot_keys)``.

    plot_labels : list (optional)
        Labels for the z-axis. Default is the plot key.
        ``len(plot_labels) == len(plot_keys)``.

    num_cols : int (optional)
        The number of columns in the resulting plot grid. Default is 1.

    fig_size : numeric (optional)
        The size of an individual subplot in inches. Default is 3.

    plot_lmfits : bool (optional)
        Whether or not to show a fit. The fit must exist! Default is False.
        Deprecated! Use plot_fits.

    plot_fits : list (optional)
        A list of fit keys to plot. Fits must exist for all parameters specified
        in plot_keys. Example: `plot_fits = ['lmfit', 'lmfit_joint_f0+qi']`
        will plot the `lmfit` results for each key in plot_keys, and the joint
        fit results from the 'f0+qi' fit for each key in plot_keys. If only one
        fit result is desired, a single string may be passed instead of a list.

    plot_kwargs : dict (optional)
        A dictionary of keyword arguments to pass the plotting function.
        Default is ``None``.

    Note
    ----
    This function is currently a little buggy, because I can't figure out how
    to intelligently adjust the label positions, sizes, etc to deal with large
    numbers in the ticks. The current workaround is to pick a large fig_size
    (so far anything larger than 5 seems ok) and then scale the plot as needed
    in some other application.

    You can also use something like the following to adjust the ticks on a
    specific axis::

        figX.axes[0].tick_params(axis='z', pad=8)
        figX.axes[0].zaxis.labelpad = 13

    """
    #TODO: Fix the labels issues when tick labels are very long.

    #Some plotting niceties
    plot_labels = kwargs.pop('plot_labels', None)
    if plot_labels is not None:
        assert len(plot_labels) == len(plot_keys), "Number of labels must equal number of plots."

    unit_multipliers = kwargs.pop('unit_multipliers', None)
    if unit_multipliers is not None:
        assert len(unit_multipliers) == len(plot_keys), "One multiplier per plot required."

    #Set some limits
    min_temp = kwargs.pop('min_temp', min(resSweep.tvec))
    max_temp = kwargs.pop('max_temp', max(resSweep.tvec))
    t_filter = (resSweep.tvec >= min_temp) * (resSweep.tvec <= max_temp)

    min_pwr = kwargs.pop('min_pwr', min(resSweep.pvec))
    max_pwr = kwargs.pop('max_pwr', max(resSweep.pvec))
    p_filter = (resSweep.pvec >= min_pwr) * (resSweep.pvec <= max_pwr)

    #Check for fits

    #plot_lmfits is deprecated and due to be removed
    plot_lmfits = kwargs.pop('plot_lmfits', False)

    if plot_lmfits:
        assert all('lmfit_'+key in resSweep.keys() for key in plot_keys), "No fit to plot for "+key+"."

    #get list of fits to plot
    plot_fits = kwargs.pop('plot_fits', None)

    #Check that all the requested plots exist
    if plot_fits is not None:
        requested_fits = []
        #Hacky way to turn a single string into a list
        if len(np.shape(plot_fits)) == 0:
            plot_fits = [plot_fits]
        for fit in plot_fits:
            for key in plot_keys:
                requested_fits.append(fit + '_' + key)

        assert all(fit in resSweep.keys() for fit in requested_fits), "No fit to plot for "+fit+"."

    #Get all the possible temperature/power combos in two lists
    ts, ps = np.meshgrid(resSweep.tvec[t_filter], resSweep.pvec[p_filter])

    #Set up the figure
    figS = plt.figure()

    fig_size = kwargs.pop('fig_size', 3)

    assert all(key in resSweep.keys() for key in plot_keys), "Unknown key"

    num_keys = len(plot_keys)
    num_cols = kwargs.pop('num_cols', 1)

    #Don't need more columns than plots
    if num_keys < num_cols:
        num_cols = num_keys

    #Calculate rows for figure size
    num_rows = int(np.ceil(1.0*num_keys/num_cols))

    #Set figure size, including some extra spacing for the colorbar
    figS.set_size_inches(fig_size*(num_cols)*1.5, fig_size*num_rows)

    #Initialize the grid for plotting
    plt_grid = gs.GridSpec(num_rows, num_cols)

    #Grab any kwargs for the plotting functions
    plot_kwargs = kwargs.pop('plot_kwargs', {})

    #Set some defaults that can be overwritten
    if 'cstride' not in plot_kwargs.keys():
        plot_kwargs['cstride'] = 1

    if 'rstride' not in plot_kwargs.keys():
        plot_kwargs['rstride'] = 1

    if 'cmap' not in plot_kwargs.keys():
        plot_kwargs['cmap'] = 'viridis'

    if 'alpha' not in plot_kwargs.keys():
        plot_kwargs['alpha'] = 0.2

    #Grab any kwargs for the fits
    fit_kwargs = kwargs.pop('fit_kwargs', {})

    #Set some defaults
    if 'color' not in fit_kwargs.keys():
        fit_kwargs['color'] = 'k'

    if 'linestyle' not in fit_kwargs.keys():
        fit_kwargs['linestyle'] ='--'



    #Loop through all the keys in the ResonatorSweep object and plot them
    for ix, key in enumerate(plot_keys):

        iRow = int(ix/num_cols)
        iCol = ix%num_cols

        axs = figS.add_subplot(plt_grid[iRow, iCol], projection='3d')

        axs.ticklabel_format(useOffset=False)

        plt_data = resSweep[key].loc[t_filter, p_filter].values.T

        if unit_multipliers is not None:
            mult = unit_multipliers[ix]
        else:
            mult = 1

        axs.plot_surface(ts, ps, mult*plt_data, **plot_kwargs)

        if plot_lmfits:
            fit_data = resSweep['lmfit_'+key].loc[t_filter, p_filter].values.T
            axs.plot_wireframe(ts, ps, mult*fit_data, **fit_kwargs)

        #Deprecated. Will be removed in next major version
        if plot_fits is not None:
            for fit in plot_fits:
                fit_data = resSweep[fit + '_' + key].loc[t_filter, p_filter].values.T
                axs.plot_wireframe(ts, ps, mult*fit_data, **fit_kwargs)

        #TODO: Implement plot_fits here.

        if plot_labels is not None:
            axs.set_zlabel(plot_labels[ix])
        else:
            axs.set_zlabel(key)
        axs.set_ylabel('Power (dB)')
        axs.set_xlabel('Temperature (mK)')


    figS.tight_layout()

    return figS



from __future__ import division
import pandas as pd
import numpy as np
import lmfit as lf
from .resonator import makeResFromData, makeResList, indexResList
from .process_file import process_file

#This is a glorified dictionary with a custom initalize method. It takes a list
#of resonator objects that have been fit, and then sets up a
#dict of pandas DataFrame objects, one for each interesting fit parameter
#This adds no new information, but makes accessing the fit data easier
class ResonatorSweep(dict):
    r"""Dictionary object with custom ``__init__`` method.\

    Parameters
    ----------
    resList : list-like
        A list of ``scraps.Resonator`` objects. Each object must have the
        attribute ``hasFit == True``.

    Keyword Arguments
    -----------------
    label : string (optional)
        Selects which set of fit data to build the ResonatorSweep from. Default is 'default'.

    index : string{'raw', 'round', 'block'} (optional)
        Selects which method to use for indexing.

    roundto : int (optional)
        Number to round temperature index to in mK. Default is 5.

    Attributes
    ----------
    tvec : array-like[nUniqeTemps]
        Index of temperature values, one for each unique temperature.

    pvec : array-like[nUniqePowers]
        Index of powers values, one for each unique power.

    smartindex : string
        Indicates the method of formatting temperature values in tvec. 'raw'
        means to take temperature values as they are, 'round' means to round to
        the neareast `X` mK, where `X` is set by `roundTo`, 'block' means to
        figure out which temperature values are nominally the same and set each
        block of those temperatures to the same value.

    rountTo : float
        The number of mK to round to when ``smartIndex == 'round'``.

    lmfit_results : dict
        A dictionary containing the fit results for different data by key.
        Initially empty, results are added by calling
        ``ResonatorSweep.do_lmfit``.

    emcee_results : dict
        A dictionary containing the MCMC results for different data by key.
        Initially empty, results are added by calling
        ``ResonatorSweep.do_emcee``.

    lmfit_joint_results : dict
        If multiple data sets are fit simultaneously via
        ``ResonatorSweep.do_lmfit``, the results will appear in this attribute
        by key, where ``key = 'key1+key2' == 'key2+key1'.

    emcee_joint_results : dict
        If multiple data sets are fit simultaneously via
        ``ResonatorSweep.do_emcee``, the results will appear in this attribute
        by key, where ``key = 'key1+key2'.

    Note
    ----
    The following keys are added to the self dict:

    'temps' : ``pandas.DataFrame``
        The temperature of each resonator in the sweep.

    'fmin' : ``pandas.DataFrame``
        The minimum value of the magnitude vs frequency curve after subtraction
        of the best-guess baseline.

    'listIndex' : ``pandas.DataFrame``
        The sweep objects are built from a list of ``pyres.Resonator``
        objects; this is the index of the original list for
        each data value in the sweep.

    If the ``do_lmfit`` method has been run, these keys are also added:

    'chisq' : ``pandas.DataFrame``
        The Chi-squared value of each fit in the sweep.

    'redchi' : ``pandas.DataFrame``
        The reduced Chi-squared value of each fit in the sweep.

    'feval' : ``pandas.DataFrame``
        The number of function evaluations for each fit in the sweep.

    paramNames : ``pandas.DataFrame``
        There is a key for each parameter value in the ``Resonator.params``
        attribute.

    paramNames + '_sigma' : ``pandas.DataFrame``
        There is a key for the uncertainty on each paramter value returned by
        the fit. Key name is paramName + '_sigma'

    If the ``do_emcee`` method has been run, these keys are also added:

    paramNames + '_mle' : ``pandas.DataFrame``
        The maximum-liklihood estimate from the MCMC sampling.

    paramNames + '_mc' : ``pandas.DataFrame``
        The 50th percentile value of the posterior distribution from MCMC
        sampling.

    paramNames + '_sigma_plus_mc' : ``pandas.DataFrame``
        The plus errorbar on the MCMC data. Calculated as the difference between
        the 84th percentile value and the 50th percentile value.

    paramNames + '_sigma_minus_mc' : ``pandas.DataFrame``
        The minus errorbar on the MCMC data. Calculated as the difference between
        the 50th percentile value and the 16th percentile value.


    """


    def __init__(self, resList, **kwargs):
        """Formats temp/pwr sweeps into easily parsed pandas DataFrame objects.

        """
        #Call the base class initialization for an empty dict.
        #Not sure this is totally necessary, but don't want to break the dict...
        dict.__init__(self)

        #Create some objects that will be filled in the future:
        self.lmfit_results = {} #Holds fit results for individual quantities
        self.emcee_results = {}
        self.lmfit_joint_results = {} #Holds fit results for joint quantities
        self.emcee_joint_results = {}

        #Build a list of keys that will eventually become the dict keys:

        #Start with the list of fit parameters, want to save all of them
        #Can just use the first resonator's list, as they are all the same.
        #params is NOT an lmfit object.
        params = list(resList[0].params.keys())

        #Add a few more
        #TODO: Right now, only fit information from the most recent fit is stored
        #TODO: Would be good to maybe have keys for each joint fit?
        params.append('temps') #Actual temperature value of measured resonator
        params.append('fmin') #Frequency at magnitude minimum
        params.append('chisq') #Chi-squared value from fit
        params.append('redchi') #Reduced chi-squared value
        params.append('feval') #Number of function evaluations to converge on fit
        params.append('listIndex') #Index in resonator list of resonator

        #If a model explicitly uses qi and qc, then calculate q0
        if all(p in params for p in ['qi', 'qc']):
            params.append('q0') #The total Q = qi*qc/(qi+qc)


        #This flag sets which data to pull from the Resonator objects
        self.label = kwargs.pop('label', 'default')
        assert all([(self.label in res.lmfit_result.keys()) for res in resList]), "label must be present in lmfit_result dict keys for every Resonator in resList."

        #This flag sets different indexing methods
        self.smartindex = kwargs.pop('index', 'raw')
        assert self.smartindex in ['raw', 'round', 'block'], "index must be 'raw', 'round', or 'block'."

        #Default the rounding option to 5 mK
        self.roundto = kwargs.pop('roundto', 5)

        if kwargs:
            raise ValueError("Unknown keyword: " + kwargs.keys()[0])


        #Loop through the resList and make lists of power and index temperature
        tvals = np.empty(len(resList))
        pvals = np.empty(len(resList))

        if self.smartindex == 'round':
            itvals = np.empty(len(resList))

        for index, res in enumerate(resList):
            tvals[index] = res.temp
            pvals[index] = res.pwr

            if self.smartindex == 'round':
                #itemp is stored in mK
                itmp = np.round(res.temp*1000/self.roundto)*self.roundto
                itvals[index] = itmp


        #Create index vectors containing only the unique values from each list
        tvec = np.sort(np.unique(tvals))
        self.pvec = np.sort(np.unique(pvals))

        #Because of uncertainty and fluctuation in temperature measurements,
        #not every temperature value / power value combination has data.
        #We want to assign index values in a smart way to get rid of empty combinations


        #Check to make sure that there aren't any rogue extra points that will mess this up
        if (self.smartindex == 'block') and (len(resList) % len(self.pvec) == 0) and (len(resList)>0):
            temptvec = [] #Will add to this as we find good index values

            tindex = 0
            setindices = []
            settemps = []
            for temp in tvec:
                for pwr in self.pvec:
                    curindex = indexResList(resList, temp, pwr)
                    if curindex is not None:
                        setindices.append(curindex)
                        settemps.append(temp)

                if len(setindices) % len(self.pvec) == 0:
                    #For now this switches to mK instead of K because of the
                    #stupid way python handles float division (small errors)
                    itemp = np.round(np.mean(np.asarray(settemps))*1000)
                    temptvec.append(itemp)

                    #Set the indexing temperature of the resonator object
                    for index in setindices:
                        resList[index].itemp = itemp

                    setindices = []
                    settemps = []

            self.tvec = np.asarray(temptvec)

            assert all(hasattr(res, 'itemp') for res in resList), "Something went wrong with assigning itemps for block mode. Make sure data is square!"

        elif self.smartindex == 'raw':
            for res in resList:
                res.itemp = np.round(res.temp*1000)
            self.tvec = np.round(tvec*1000)
        elif self.smartindex == 'round':
            for index, res in enumerate(resList):
                res.itemp = itvals[index]
            self.tvec = np.sort(np.unique(itvals))
        else:
            self.tvec = np.round(tvec*1000)
            self.smartindex = 'raw'
            for res in resList:
                res.itemp = np.round(res.temp*1000)

        
        #Loop through the parameters list and create a DataFrame for each one
        for pname in params:
            #Start out with a 2D dataframe full of NaN of type float
            #Row and Column indices are temperature and power values
            self[pname] = pd.DataFrame(np.nan, index = self.tvec, columns = self.pvec)


            if pname in resList[0].params.keys():
                #Uncertainty on best fit from least-squares
                self[pname+'_sigma'] = pd.DataFrame(np.nan, index=self.tvec, columns = self.pvec)

                #Maximum liklihood value from MCMC
                self[pname+'_mle'] = pd.DataFrame(np.nan, index = self.tvec, columns = self.pvec)

                #50th percentile value of MCMC chain
                self[pname+'_mc'] = pd.DataFrame(np.nan, index = self.tvec, columns = self.pvec)

                #84th-50th values from MCMC chain
                self[pname+'_sigma_plus_mc'] = pd.DataFrame(np.nan, index = self.tvec, columns = self.pvec)

                #50th-16th values from MCMC chain
                self[pname+'_sigma_minus_mc'] = pd.DataFrame(np.nan, index = self.tvec, columns = self.pvec)


            #Fill it with as much data as exists
            for index, res in enumerate(resList):
                if pname in res.lmfit_result[self.label]['result'].params.keys():
                    if res.lmfit_result[self.label]['result'].params[pname].vary is True:
                        #The actual best fit value
                        self[pname][res.pwr][res.itemp] = res.lmfit_result[self.label]['result'].params[pname].value

                        #Get the right index to find the uncertainty in the covariance matrix
                        cx = res.lmfit_result[self.label]['result'].var_names.index(pname)

                        #The uncertainty is the sqrt of the autocovariance
                        if res.lmfit_result[self.label]['result'].covar is not None:
                            self[pname+'_sigma'][res.pwr][res.itemp] = np.sqrt(res.lmfit_result[self.label]['result'].covar[cx, cx])

                        #Get the maximum liklihood if it exists
                        if res.hasChain is True:

                            #Grab the index of the parameter in question
                            sx = list(res.emcee_result[self.label]['result'].flatchain.iloc[np.argmax(res.emcee_result[self.label]['result'].lnprob)].keys()).index(pname)
                            self[pname+'_mle'][res.pwr][res.itemp] = res.emcee_result[self.label]['mle_vals'][pname]
                            self[pname+'_mc'][res.pwr][res.itemp] = res.emcee_result[self.label]['result'].params[pname].value

                            #Since the plus and minus errorbars can be different,
                            #have to store them separately
                            self[pname+'_sigma_plus_mc'][res.pwr][res.itemp] = res.emcee_result[self.label]['emcee_sigmas'][sx][0]
                            self[pname+'_sigma_minus_mc'][res.pwr][res.itemp] = res.emcee_result[self.label]['emcee_sigmas'][sx][1]
                elif pname == 'temps':
                    #Since we bin the temps by itemp for indexing, store the actual temp here
                    self[pname][res.pwr][res.itemp] = res.temp
                elif pname == 'fmin':
                    self[pname][res.pwr][res.itemp] = res.fmin
                elif pname == 'chisq':
                    self[pname][res.pwr][res.itemp] = res.lmfit_result[self.label]['result'].chisqr
                elif pname == 'redchi':
                    self[pname][res.pwr][res.itemp] = res.lmfit_result[self.label]['result'].redchi
                elif pname == 'feval':
                    self[pname][res.pwr][res.itemp] = res.lmfit_result[self.label]['result'].nfev
                elif pname == 'listIndex':
                    #This is useful for figuring out where in the resList the data you care about is
                    self[pname][res.pwr][res.itemp] = index
                elif pname == 'q0':
                    qi = res.lmfit_result[self.label]['result'].params['qi'].value
                    qc = res.lmfit_result[self.label]['result'].params['qc'].value
                    self[pname][res.pwr][res.itemp] = qi*qc/(qi+qc)

    def do_lmfit(self, fit_keys, models_list, params_list, model_kwargs=None, param_kwargs=None, **kwargs):
        r"""Run simulatneous fits on the temp/pwr data for several parameters.
        Results are stored in either the ``lmfit_results`` or
        ``lmfit_joint_results`` attribute depending on whether one or multiple
        keys are passed to `fit_keys`.

        Parameters
        ----------
        fit_keys : list-like
            A list of keys that correspond to existing data. Any combination of
            keys from `self.keys()`` is acceptable, but duplicates are not
            permitted.

        models_list : list-like
            A list of fit functions, one per key in `fit_keys`. Function must
            return a residual of the form: ``residual = (model-data)/sigma``
            where ``residual``, ``model``, and ``data`` are all ``numpy``
            arrays. Function signature is ``model_func(params, temps, powers,
            data=None, sigmas=None)``. If ``data==None`` the functions must
            return the model calculated at ``temps`` and ``powers``. The model
            functions should also gracefully handle ``np.NaN`` or ``None``
            values.

        params_list : list-like
            A list of ``lmfit.Parameters`` objects, one for each key in
            `fit_keys`. Parameters sharing the same name will be merged so that
            the fit is truly joint. Alternately, a list of functions that return
            ``lmfit.Parameters`` objects may be passed. In this case, one should
            use `param_kwargs` to pass any needed options to the functions.

        model_kwargs : list-like (optional)
            A list of ``dict`` objects to pass to the individual model functions
            as kwargs. ``None`` is also an acceptable entry  if there are no
            kwargs to pass to a model function. Default is ``None.``

        param_kwargs : list-like (optional)
            A list of ``dict`` objects to pass to the individual params
            functions as kwargs. ``None`` is also an acceptable entry  if
            there are no kwargs to pass to a model function. Default is
            ``None.``

        lmfit_kwargs : dict (optional)
            Keyword arguments to pass options to the fitter

        kwargs : dict (optional)
            Supported keyword arugments are 'min_temp', 'max_temp', 'min_pwr',
            and 'max_pwr'. These set limits on which data to fit.

        Keyword arguments
        -----------------

        min_temp : numeric
            Lower limit of temperature to fit. Default is 0.

        max_temp : numeric
            Upper limit of temerature to fit. Default is infinity.

        min_pwr : numeric
            Lower limit of temperature to fit. Default is -infinity.

        max_pwr : numeric
            Upper limit of temperature to fit. Default is infinity.

        raw_data : string {'lmfit', 'emcee', 'mle'}
            Whether to use the values returned by lmfit, or the values returned
            by the emcee fitter (either the 50th percentile or the maximum
            liklihood). This also chooses which set of errorbars to use: either
            those from the lmfit covariance matrix, or those from the 16th and
            84th percentiles of the posterior probablility distribution. Default
            is 'lmfit'.

        Note
        ----
        If the fits are succesful, the resulting fit data (ie the best fit
        surface) will be added to the self dict in the form of a
        ``pandas.DataFrame`` under the following keys:

        For a joint fit (``len(fit_keys) > 1``)::

            'lmfit_joint_'+joint_key+'_'+key for each key in fit_keys

        For a single fit (``len(fit_keys) == 1``)::

            'lmfit_'+key

        """



        #Set some limits
        min_temp = kwargs.pop('min_temp', min(self.tvec))
        max_temp = kwargs.pop('max_temp', max(self.tvec))
        t_filter = (self.tvec >= min_temp) * (self.tvec <= max_temp)

        min_pwr = kwargs.pop('min_pwr', min(self.pvec))
        max_pwr = kwargs.pop('max_pwr', max(self.pvec))
        p_filter = (self.pvec >= min_pwr) * (self.pvec <= max_pwr)

        #Process the final kwarg:
        raw_data = kwargs.pop('raw_data', 'lmfit')
        assert raw_data in ['lmfit', 'emcee', 'mle'], "raw_data must be 'lmfit' or 'emcee'."



        assert len(fit_keys) == len(models_list) == len(params_list), "Make sure argument lists match in number."

        #Make some empty dictionaries just in case
        if model_kwargs is None:
            model_kwargs = [{}]*len(fit_keys)

        if param_kwargs is None:
            params_kwargs = [{}]*len(fit_keys)


        #Check to see if this should go in the joint_fits dict, and build a key if needed.
        if len(fit_keys) > 1:
            joint_key = '+'.join(fit_keys)
        else:
            joint_key = None

        #Check if params looks like a lmfit.Parameters object.
        #If not, assume is function and try to set params by calling it
        for px, p in enumerate(params_list):
            if not hasattr(p, 'valuesdict'):
                assert params_kwargs[px] is not None, "If passing functions to params, must specfify params_kwargs."
                params_list[px] = p(**param_kwargs[px])

        #Combine the different params objects into one large list
        #Only the first of any duplicates will be transferred
        merged_params = lf.Parameters()
        if len(params_list) > 1:
            for p in params_list:
                for key in p.keys():
                    if key not in merged_params.keys():
                        merged_params[key] = p[key]
        else:
            merged_params = params_list[0]

        #Get all the possible temperature/power combos into two grids
        ts, ps = np.meshgrid(self.tvec[t_filter], self.pvec[p_filter])

        #Create grids to hold the fit data and the sigmas
        fit_data_list = []
        fit_sigmas_list = []

        #Get the data that corresponds to each temperature power combo and
        #flatten it to match the ts/ps combinations
        #Transposing is important because numpy matrices are transposed from
        #Pandas DataFrames
        for key in fit_keys:

            if raw_data == 'emcee':
                key = key + '_mc'
            elif raw_data == 'mle':
                key = key + '_mle'

            if raw_data in ['emcee', 'mle']:
                err_bars = (self[key+'_sigma_plus_mc'].loc[t_filter, p_filter].values.T+
                            self[key+'_sigma_minus_mc'].loc[t_filter, p_filter].values.T)
            else:
                err_bars = self[key+'_sigma'].loc[t_filter, p_filter].values.T

            fit_data_list.append(self[key].loc[t_filter, p_filter].values.T)
            fit_sigmas_list.append(err_bars)

        #Create a new model function that will be passed to the minimizer.
        #Basically this runs each fit and passes all the residuals back out
        def model_func(params, models, ts, ps, data, sigmas, kwargs):
            residuals = []
            for ix, key in enumerate(fit_keys):
                residuals.append(models[ix](params, ts, ps, data[ix], sigmas[ix], **kwargs[ix]))

            return np.asarray(residuals).flatten()


        #Create a lmfit minimizer object
        minObj = lf.Minimizer(model_func, merged_params, fcn_args=(models_list, ts, ps, fit_data_list, fit_sigmas_list, model_kwargs))

        #Call the lmfit minimizer method and minimize the residual
        lmfit_result = minObj.minimize(method = 'leastsq')

        #Put the result in the appropriate dictionary
        if joint_key is not None:
            self.lmfit_joint_results[joint_key] = lmfit_result
        else:
            self.lmfit_results[fit_keys[0]] = lmfit_result

        #Calculate the best-fit model from the params returned
        #And put it into a pandas DF with the appropriate key.
        #The appropriate key format is: 'lmfit_joint_'+joint_key+'_'+key
        #or, for a single fit: 'lmfit_'+key
        for ix, key in enumerate(fit_keys):
            #Call the fit model without data to have it return the model
            returned_model = models_list[ix](lmfit_result.params, ts, ps)

            #Build the appropriate key
            if joint_key is not None:
                new_key = 'lmfit_joint_'+joint_key+'_'+key
            else:
                new_key = 'lmfit_'+key

            #Make a new dict entry to the self dictioary with the right key.
            #Have to transpose the matrix to turn it back into a DF
            self[new_key] = pd.DataFrame(np.nan, index=self.tvec, columns=self.pvec)
            self[new_key].loc[self.tvec[t_filter], self.pvec[p_filter]] = returned_model.T




    def do_emcee(self, fit_keys, models_list, params_list=None, model_kwargs=None, param_kwargs=None, emcee_kwargs=None, **kwargs):
        r"""Run simulatneous MCMC sampling on the temp/pwr data for several
        parameters. Results are stored in either the ``emcee_results`` or
        ``emcee_joint_results`` attribute depending on whether one or multiple
        keys are passed to `fit_keys`.

        Parameters
        ----------
        fit_keys : list-like
            A list of keys that correspond to existing data. Any combination of
            keys from `self.keys()`` is acceptable, but duplicates are not
            permitted.

        models_list : list-like
            A list of fit functions, one per key in `fit_keys`. Function must
            return a residual of the form: ``residual = (model-data)/sigma``
            where ``residual``, ``model``, and ``data`` are all ``numpy``
            arrays. Function signature is ``model_func(params, temps, powers,
            data=None, sigmas=None)``. If ``data==None`` the functions must
            return the model calculated at ``temps`` and ``powers``. The model
            functions should also gracefully handle ``np.NaN`` or ``None``
            values.

        params_list : list-like
            A list of ``lmfit.Parameters`` objects, one for each key in
            `fit_keys`. Parameters sharing the same name will be merged so that
            the fit is truly joint. Alternately, a list of functions that return
            ``lmfit.Parameters`` objects may be passed. In this case, one should
            use `param_kwargs` to pass any needed options to the functions.
            Default is ``None`` and is equivalent to setting ``use_lmfit_params =
            True``.

        model_kwargs : list-like (optional)
            A list of ``dict`` objects to pass to the individual model functions
            as kwargs. ``None`` is also an acceptable entry  if there are no
            kwargs to pass to a model function. Default is ``None.``

        param_kwargs : list-like (optional)
            A list of ``dict`` objects to pass to the individual params
            functions as kwargs. ``None`` is also an acceptable entry  if
            there are no kwargs to pass to a model function. Default is
            ``None.``

        emcee_kwargs : dict (optional)
            Keyword arguments to pass options to the fitter

        Keyword Arguments
        -----------------
        min_temp : numeric
            Lower limit of temperature to fit. Default is 0.

        max_temp : numeric
            Upper limit of temerature to fit. Default is infinity.

        min_pwr : numeric
            Lower limit of temperature to fit. Default is -infinity.

        max_pwr : numeric
            Upper limit of temperature to fit. Default is infinity.

        use_lmfit_params : bool
            Whether or not to use the resulting best-fit ``lmfit.Paramters``
            object that resulted from calling ``ResonatorSweep.do_lmfit()`` as
            the starting value for the MCMC sampler. Default is True.

        raw_data : string {'lmfit', 'emcee', 'mle'}
            Whether to use the values returned by lmfit, or the values returned
            by the emcee fitter (either the 50th percentile or the maximum
            liklihood). This also chooses which set of errorbars to use: either
            those from the lmfit covariance matrix, or those from the 16th and
            84th percentiles of the posterior probablility distribution. Default
            is 'lmfit'.

        Note
        ----
        If the fits are succesful, the resulting fit data (ie the best fit
        surface) will be added to the self dict in the form of a
        ``pandas.DataFrame`` under the following keys:

        For a joint fit (``len(fit_keys) > 1``)::

            'emcee_joint_'+joint_key+'_'+key for each key in fit_keys

        For a single fit (``len(fit_keys) == 1``)::

            'emcee_'+key

        """

        #Figure out which data to fit
        raw_data = kwargs.pop('raw_data', 'lmfit')
        assert raw_data in ['lmfit', 'emcee', 'mle'], "raw_data must be 'lmfit' or 'emcee'."



        #Set some limits
        min_temp = kwargs.pop('min_temp', min(self.tvec))
        max_temp = kwargs.pop('max_temp', max(self.tvec))
        t_filter = (self.tvec >= min_temp) * (self.tvec <= max_temp)

        min_pwr = kwargs.pop('min_pwr', min(self.pvec))
        max_pwr = kwargs.pop('max_pwr', max(self.pvec))
        p_filter = (self.pvec >= min_pwr) * (self.pvec <= max_pwr)



        if params_list is not None:
            assert len(fit_keys) == len(models_list) == len(params_list), "Make sure argument lists match in number."
        else:
            assert len(fit_keys) == len(models_list), "Make sure argument lists match in number."

        #Make some empty dictionaries just in case so we don't break functions
        #by passing None as a kwargs
        if model_kwargs is None:
            model_kwargs = [{}]*len(fit_keys)

        if param_kwargs is None:
            params_kwargs = [{}]*len(fit_keys)

        if emcee_kwargs is None:
            emcee_kwargs = {}


        #Check to see if this should go in the joint_fits dict, and build a key if needed.
        if len(fit_keys) > 1:
            joint_key = '+'.join(fit_keys)
        else:
            joint_key = None


        #If possible (and desired) then we should use the existing best fit as a starting point
        #For the MCMC sampling. If not, build params from whatever is passed in.
        use_lmfit_params = kwargs.pop('use_lmfit_params', True)

        if (params_list is not None) and (use_lmfit_params == False):

            #Check if params looks like a lmfit.Parameters object.
            #If not, assume is function and try to set params by calling it
            for px, p in enumerate(params_list):
                if not hasattr(p, 'valuesdict'):
                    assert params_kwargs[px] is not None, "If passing functions to params, must specfify params_kwargs."
                    params_list[px] = p(**param_kwargs[px])

            #Combine the different params objects into one large list
            #Only the first of any duplicates will be transferred
            merged_params = lf.Parameters()
            if len(params_list) > 1:
                for p in params_list:
                    for key in p.keys():
                        if key not in merged_params.keys():
                            merged_params[key] = p[key]
            else:
                merged_params = params_list[0]

        else:
            if joint_key is not None:
                assert joint_key in self.lmfit_joint_results.keys(), "Can't use lmfit params. They don't exist."
                merged_params = self.lmfit_joint_results[joint_key].params
            else:
                assert fit_keys[0] in self.lmfit_results.keys(), "Can't use lmfit params. They don't exist."
                merged_params = self.lmfit_results[fit_keys[0]].params


        #Get all the possible temperature/power combos into two grids
        ts, ps = np.meshgrid(self.tvec[t_filter], self.pvec[p_filter])

        #Create grids to hold the fit data and the sigmas
        fit_data_list = []
        fit_sigmas_list = []

        #Get the data that corresponds to each temperature power combo and
        #flatten it to match the ts/ps combinations
        #Transposing is important because numpy matrices are transposed from
        #Pandas DataFrames
        for key in fit_keys:

            if raw_data == 'emcee':
                key = key + '_mc'
            elif raw_data == 'mle':
                key = key + '_mle'

            if raw_data in ['emcee', 'mle']:
                err_bars = (self[key+'_sigma_plus_mc'].loc[t_filter, p_filter].values.T+
                            self[key+'_sigma_minus_mc'].loc[t_filter, p_filter].values.T)
            else:
                err_bars = self[key+'_sigma'].loc[t_filter, p_filter].values.T

            fit_data_list.append(self[key].loc[t_filter, p_filter].values.T)
            fit_sigmas_list.append(err_bars)

        #Create a new model function that will be passed to the minimizer.
        #Basically this runs each fit and passes all the residuals back out
        def model_func(params, models, ts, ps, data, sigmas, kwargs):
            residuals = []
            for ix, key in enumerate(fit_keys):
                residuals.append(models[ix](params, ts, ps, data[ix], sigmas[ix], **kwargs[ix]))

            return np.asarray(residuals).flatten()


        #Create a lmfit minimizer object
        minObj = lf.Minimizer(model_func, merged_params, fcn_args=(models_list, ts, ps, fit_data_list, fit_sigmas_list, model_kwargs))

        #Call the lmfit minimizer method and minimize the residual
        emcee_result = minObj.emcee(**emcee_kwargs)

        #Put the result in the appropriate dictionary
        if joint_key is not None:
            self.emcee_joint_results[joint_key] = emcee_result
        else:
            self.emcee_results[fit_keys[0]] = emcee_result

        #Calculate the best-fit model from the params returned
        #And put it into a pandas DF with the appropriate key.
        #The appropriate key format is: 'lmfit_joint_'+joint_key+'_'+key
        #or, for a single fit: 'lmfit_'+key
        for ix, key in enumerate(fit_keys):
            #Call the fit model without data to have it return the model
            returned_model = models_list[ix](emcee_result.params, ts, ps)

            #Build the appropriate key
            if joint_key is not None:
                new_key = 'emcee_joint_'+joint_key+'_'+key
            else:
                new_key = 'emcee_'+key

            #Make a new dict entry to the self dictioary with the right key.
            #Have to transpose the matrix to turn it back into a DF
            self[new_key] = pd.DataFrame(np.nan, index=self.tvec, columns=self.pvec)
            self[new_key].loc[self.tvec[t_filter], self.pvec[p_filter]] = returned_model.T

    def info():
        """Print out some information on all the keys that are stored in the object."""

        #For now, this just spits out all the keys. Could be more useful.
        print(sorted(self.keys()))



from .resonator import Resonator, makeResFromData, makeResList, indexResList
from .resonator_sweep import ResonatorSweep
from .fitsS21 import cmplxIQ_fit, cmplxIQ_params
from .fitsSweep import qi_tlsAndMBT, f0_tlsAndMBT
from .process_file import process_file
from .plot_tools import plotResListData, plotResSweepParamsVsX, plotResSweepParamsVsTemp, plotResSweepParamsVsPwr, plotResSweep3D



import numpy as np
from scipy.special import digamma
import scipy.constants as sc

import barmat


def qi_tlsAndMBT(params, temps, powers, data=None, eps=None, **kwargs):
    """A model of internal quality factor vs temperature and power, weighted by uncertainties.

    Parameters
    ----------
    params : ``lmfit.Parameters`` object
        Parameters must include ``['Fd', 'q0', 'f0', 'alpha', 'tc', 'vf',
        'london0', 'mfp', 'bcs']``.

    temps : ``numpy.Array``
        Array of temperature values to evaluate model at. May be 2D.

    powers : ``numpy.Array``
        Array of power values to evaluate model at. May be 2D.

    data : ``numpy.Array``
        Data values to compare to model. May also be ``None``, in which case
        function returns model.

    eps : ``numpy.Array``
        Uncertianties with which to weight residual. May also be ``None``, in
        which case residual is unwieghted.

    Returns
    -------

    residual : ``numpy.Array``
        The weighted or unweighted vector of residuals if ``data`` is passed.
        Otherwise, it returns the model.

    Note
    ----
    The following constraint must be satisfied::

        all(numpy.shape(x) == numpy.shape(data) for x in [temps, powers, eps])

    This model does a complete numerical calculation of the Mattis-Bardeen
    surface impedance.

    Some equations are taken from J. Gao's Caltech dissertation (2008):

    (2.54) gives for MBD: ``1/Q(T)-1/Q(0) = alpha * R(T)/X(0)``

    (5.72) and (5.65) give for TLS: ``1/Q(T)-1/Q(0) = Fd*tanh(hf/2kT)/sqrt(1+P/P0)``

    R(T)/X(0) is calculated using the barmat python package at:
    http://github.com/FaustinCarter/barmat.

    """

    #TLS params
    Fd = params['Fd'].value
    Pc = params['Pc'].value
    f0 = params['f0'].value
    q0 = params['q0'].value

    #Barmat params
    tc = params['tc'].value
    vf = params['vf'].value
    london0 = params['london0'].value
    mfp = params['mfp'].value
    bcs = params['bcs'].value

    #Kinetic inductance param
    alpha = params['alpha'].value

    units = kwargs.pop('units', 'mK')
    assert units in ['mK', 'K'], "Units must be 'mK' or 'K'."

    if units == 'mK':
        temps = temps*0.001

    #Pack all these together for convenience
    zeta = sc.h*f0/(2*sc.k*temps)

    #An optional power calibration in dB
    #Without this, the parameter Pc is meaningless
    pwr_cal_dB = kwargs.pop('pwr_cal_dB', 0)
    ps = powers+pwr_cal_dB

    #Working in inverse Q since they add and subtract  nicely

    #Calculate the inverse Q from TLS
    invQtls = Fd*np.tanh(zeta)/np.sqrt(1.0+10**(ps/10.0)/Pc)

    #Calculate the inverse Q from MBD using barmat
    fr = sc.h*f0/(sc.e*barmat.tools.get_delta0(tc, bcs))

    Z = barmat.get_Zvec(temps[0]/tc, tc, vf, london0, mfp=mfp, bcs=bcs, fr=fr,
                        axis='temperature', output_depths=False,
                        boundary='diffuse')

    invQmbd = alpha*(Z.real - Z.real[0])/Z.imag[0]

    invQmbd = np.array([invQmbd]*len(temps))

    #Get the difference from the total Q and
    model = 1.0/(invQtls + invQmbd + 1.0/q0)

    #Weight the residual if eps is supplied
    if data is not None:
        if eps is not None:
            residual = (model-data)/eps
        else:
            residual = (model-data)

        return residual
    else:
        return model



def f0_tlsAndMBT(params, temps, powers, data = None, eps = None, **kwargs):
    """A model of frequency shift vs temperature and power, weighted by uncertainties.

    Parameters
    ----------
    params : ``lmfit.Parameters`` object
        Parameters must include ``['Fd', 'df', 'fRef', 'alpha', 'delta0']``.

    temps : ``numpy.Array``
        Array of temperature values to evaluate model at. May be 2D.

    powers : ``numpy.Array``
        Array of power values to evaluate model at. May be 2D.

    data : ``numpy.Array``
        Data values to compare to model. May also be ``None``, in which case
        function returns model.

    eps : ``numpy.Array``
        Uncertianties with which to weight residual. May also be ``None``, in
        which case residual is unwieghted.

    Returns
    -------

    residual : ``numpy.Array``
        The weighted or unweighted vector of residuals if ``data`` is passed.
        Otherwise, it returns the model.

    Note
    ----
    The following constraint must be satisfied::

        all(numpy.shape(x) == numpy.shape(data) for x in [temps, powers, eps])

    This model does a complete numerical calculation of the Mattis-Bardeen
    surface impedance. It assumes no power dependence in the frequency shift

    Some equations are taken from J. Gao's Caltech dissertation (2008):

    (2.54) gives for MBD (f(T)-f(0))/f(0) = -alpha*0.5*(X(T)-X(0))/X(0)

    (5.71) gives for TLS "" = Fd/pi * (usual TLS expression from Phillips)

    (X(T)-X(0))/X(0) calculated is calculated using the barmat python package at:
    http://github.com/FaustinCarter/barmat.

    """
    #Unpack parameter values from params

    #TLS params
    Fd = params['Fd'].value
    f0 = params['f0'].value

    #Barmat params
    tc = params['tc'].value
    vf = params['vf'].value
    london0 = params['london0'].value
    mfp = params['mfp'].value
    bcs = params['bcs'].value

    #Kinetic inductance param
    alpha = params['alpha'].value

    #Set temperature units
    units = kwargs.pop('units', 'mK')
    assert units in ['mK', 'K'], "Units must be 'mK' or 'K'."

    if units == 'mK':
        temps = temps*0.001

    #Pack all these together for convenience
    zeta = sc.h*f0/(2*sc.k*temps)

    #TLS contribution
    dfTLS = Fd/sc.pi*(np.real(digamma(0.5+zeta/(1j*sc.pi)))-np.log(zeta/sc.pi))

    #MBD contribution
    #Calculate the dfMBD from MBD using barmat
    fr = sc.h*f0/(sc.e*barmat.tools.get_delta0(tc, bcs))

    Z = barmat.get_Zvec(temps[0]/tc, tc, vf, london0, mfp=mfp, bcs=bcs, fr=fr,
                        axis='temperature', output_depths=True,
                        boundary='diffuse')

    dfMBD = -0.5*alpha*(Z.imag - Z.imag[0])/Z.imag[0]

    dfMBd = np.array([dfMBD]*len(temps))

    #Calculate model from parameters
    model = f0+f0*(dfTLS + dfMBD)


    #Weight the residual if eps is supplied
    if data is not None:
        if eps is not None:
            residual = (model-data)/eps
        else:
            residual = (model-data)

        return residual
    else:
        return model



from .resonator import Resonator, makeResFromData
import numpy as np
from astropy.io import fits

def load_fits(dataFile, resNum = 0):
    """Load in a single sweep from one of Pete's .fits files"""
    
    data = fits.open(dataFile, memmap=False)
    
    header = data[1].header
    table = data[1].data
    
    num_sweeps = header['VNATONES']
    attenuation = header['INPUTATT'] + header['ATT_UC'] + header['ATT_C'] + header['ATT_RT']
    vna_power = header['VNAPOWER']
    power = vna_power - attenuation
    
    temp = header['SAMPLETE']
    
    num_pts = header['NAXIS2']//num_sweeps
    
    dataDict = {}
    
    start_ix = resNum*num_pts
    end_ix = start_ix + num_pts - 1
    
    mask = slice(start_ix, end_ix)
    
    dataDict['freq'] = table['Freq'][mask]
    dataDict['I'] = table['ReS21'][mask]
    dataDict['Q'] = table['ImS21'][mask]
    
    dataDict['temp'] = temp
    dataDict['pwr'] = power
    dataDict['name'] = 'RES-' + str(resNum)
    
    data.close()
    
    return dataDict
    
    
    
    
    

def load_one_ESL(dataFile, sweepType='rough', resNum = 0, **kwargs):
    """Load a single resonator file from ESL into a data dict for later processing.

    Return value:
        dataDict -- a dict object with six fields
            freq, I, Q, temp, pwr, name

    Arguments:
        dataFile -- path to data
        sweepType -- 'rough', 'gain', or 'fine'. Default is 'rough'
        resNum -- the number of the resonator in the file you want to load
            indexing starts at zero

    Keyword arguments:
        'temp' -- float
        'units' -- 'mK' or 'K' depending on how you enter 'temp'
        'name' -- string. Defaults to 'RES-X'
        'legacy' -- bool swaps I and Q values if you have an old file"""
    #Set some default values
    unitString = kwargs.pop('units', 'mK')
    if unitString == 'mK':
        units = 0.001
    elif unitString == 'K':
        units = 1
    else:
        raise ValueError("Unknown unit type. Acceptable values are 'mK' or 'K'.")
    temp = kwargs.pop('temp', np.NAN)
    name = kwargs.pop('name', 'RES-X')
    legacy = kwargs.pop('legacy', False)

    if kwargs:
        raise ValueError("Unknown keyword argument passed: " + kwargs.keys()[0])

    #If you change your default sweep types, update this dict here!
    sweepDict = {'gain':201,
                'rough':401,
                'fine':1601}

    #Open file and read power from header data
    with open(dataFile) as fp:
        for i, line in enumerate(fp):
            if i == 2:
                pwr = int(line[line.find(':')+1:line.find('.')])
            elif i > 2:
                break

    #Load in the all I,Q,freq data
    cdata = np.loadtxt(dataFile, delimiter=',',skiprows=6)

    #Slice out the data from the resonator you want to look at
    numPts = sweepDict[sweepType]
    startData = resNum*numPts
    endData = (resNum+1)*numPts - 1

    #Dump everything into the dataDict
    dataDict = {}
    dataDict['freq'] = cdata[startData:endData,0]

    #Load in the IQ data
    #Old VNA programs reversed I and Q, so call legacy if trying to load those
    if legacy:
        dataDict['Q'] = cdata[startData:endData,1]
        dataDict['I'] = cdata[startData:endData,2]
    else:
        dataDict['I'] = cdata[startData:endData,1]
        dataDict['Q'] = cdata[startData:endData,2]
    ###

    dataDict['temp'] = temp*units
    dataDict['pwr'] = pwr
    dataDict['name'] = name

    return dataDict #Suitable for use with pyres.makeResFromData()



def load_sweep_ESL(dataFolder, resNames, tvals, pwrs, **kwargs):
    """Load a set of ESL resonator data corresponding to a power/temperature sweep.

    Return value:
        resList -- a dict of lists of Resonator objects, indexed by resonator name.

    Arguments:
        resNames -- a list of string identifiers. len(resNames) == number of individual resonators you scanned over
        dataFolder -- path to folder containing data, including final slash
        tvals -- a list of temperature values in the same order as the sweepXX folders
            Example: Two different temperatures
                dataFolder/
                    sweep00/ <--- 2K
                    sweep01/ <--- 100 mK

                so tval = [2., 0.1] and units='K' OR tval = [2000., 100] and units = 'mK'
        pwrs -- a list of power values in the same order as the VNAPnXX identifier in the filenames

    Keyword arguments:
        'units' -- 'mK' or 'K' depending on how you enter 'temp'
        'sweep' -- 'gain', 'rough', or 'fine'. Default is 'rough'."""
    #Set up some default values

    unitString = kwargs.pop('units', 'mK')
    if unitString == 'mK':
        units = 0.001
    elif unitString == 'K':
        units = 1
    else:
        raise ValueError("Unknown unit type. Acceptable values are 'mK' or 'K'.")


    #Change this if you change the types of sweep that exist in the program
    sweepDict = {'gain':201,
                'rough':401,
                'fine':1601}

    sweepType = kwargs.pop('sweep', 'rough')
    assert sweepType in sweepDict.keys(), "Invalid sweep type. Try 'gain', 'rough', or 'fine'."

    #If loading old data, swaps I and Q
    legacy = kwargs.pop('legacy', False)

    #Make sure no unknown keywords arguments
    if kwargs:
        raise ValueError("Unknown keyword argument passed: " + kwargs.keys()[0])

    #Declare some empty containers
    dataDicts = {}
    resLists = {}
    for resName in resNames:
        dataDicts[resName] = []
        resLists[resName] = []

    #Loop through all the directories and pull out the data
    for indexp, pwr in enumerate(pwrs):
        seriesNum = str(indexp).zfill(2)

        for indext, tval in enumerate(tvals):
            folderNum = str(indext+1).zfill(2)
            cdata = np.loadtxt(dataFolder+'sweep'+folderNum+'/'+sweepType+'VNAPn'+seriesNum+'set000.txt', delimiter=',',skiprows=6)

            for indexr, resName in enumerate(resNames):
                numPts = sweepDict[sweepType]
                startData = indexr*numPts
                endData = (indexr+1)*numPts - 1

                dataDict = {}
                dataDict['freq'] = cdata[startData:endData,0]

                if legacy:
                    dataDict['Q'] = cdata[startData:endData,1]
                    dataDict['I'] = cdata[startData:endData,2]
                else:
                    dataDict['I'] = cdata[startData:endData,1]
                    dataDict['Q'] = cdata[startData:endData,2]
                ###

                dataDict['temp'] = tval*units
                dataDict['pwr'] = pwr
                dataDict['name'] = resName

                dataDicts[resName].append(dataDict)

    #Convert the dataDicts into Resonator objects and return a dict of lists of Resonators
    for resName in resNames:
        resLists[resName] = [makeResFromData(dataDict) for dataDict in dataDicts[resName]]

    return resLists #Suitable for passing to ResonatorSweep



import numpy as np
from scipy.special import digamma, i0, k0
import scipy.constants as sc
import scipy.interpolate as si


def qi_tlsAndMBT(params, temps, powers, data=None, eps=None, **kwargs):
    """A model of internal quality factor vs temperature and power, weighted by uncertainties.

    Parameters
    ----------
    params : ``lmfit.Parameters`` object
        Parameters must include ``['Fd', 'q0', 'f0', 'alpha', 'delta0']``.

    temps : ``numpy.Array``
        Array of temperature values to evaluate model at. May be 2D.

    powers : ``numpy.Array``
        Array of power values to evaluate model at. May be 2D.

    data : ``numpy.Array``
        Data values to compare to model. May also be ``None``, in which case
        function returns model.

    eps : ``numpy.Array``
        Uncertianties with which to weight residual. May also be ``None``, in
        which case residual is unwieghted.

    Returns
    -------

    residual : ``numpy.Array``
        The weighted or unweighted vector of residuals if ``data`` is passed.
        Otherwise, it returns the model.

    Note
    ----
    The following constraint must be satisfied::

        all(numpy.shape(x) == numpy.shape(data) for x in [temps, powers, eps])

    It is almost certain that this model does NOT apply to your device as the
    assumptions it makes are highly constraining and ignore several material
    parameters. It is included here more as an example for how to write a model
    than anything else, and it does at least qualitatively describe the behavior
    of most superconducting resonators.

    This model is taken from J. Gao's Caltech dissertation (2008) and the below
    equations are from that work.

    (2.54) gives for MBD: ``1/Q(T)-1/Q(0) = alpha * R(T)/X(0)``

    (5.72) and (5.65) give for TLS: ``1/Q(T)-1/Q(0) = Fd*tanh(hf/2kT)/sqrt(1+P/P0)``

    R(T)/X(0) calculated from (2.80), (2.89), and (2.90), using the ``deltaBCS``
    function in this module for returning gap as a function of temperature.

    """

    Fd = params['Fd'].value
    Pc = params['Pc'].value
    f0 = params['f0'].value
    q0 = params['q0'].value
    delta0 = params['delta0'].value*sc.e
    alpha = params['alpha'].value

    units = kwargs.pop('units', 'mK')
    assert units in ['mK', 'K'], "Units must be 'mK' or 'K'."

    if units == 'mK':
        ts = temps*0.001

    #Assuming thick film local limit
    #Other good options are 1 or 1/3
    gamma = kwargs.pop('gamma', 0.5)

    #Calculate tc from BCS relation
    tc = delta0/(1.76*sc.k)

    #Get the reduced energy gap
    deltaR = deltaBCS(ts/tc)

    #And the energy gap at T
    deltaT = delta0*deltaR

    #Pack all these together for convenience
    zeta = sc.h*f0/(2*sc.k*ts)

    #An optional power calibration in dB
    #Without this, the parameter Pc is meaningless
    pwr_cal_dB = kwargs.pop('pwr_cal_dB', 0)
    ps = powers+pwr_cal_dB

    #Working in inverse Q since they add and subtract  nicely

    #Calculate the inverse Q from TLS
    invQtls = Fd*np.tanh(zeta)/np.sqrt(1.0+10**(ps/10.0)/Pc)

    #Calculte the inverse Q from MBD
    invQmbd = alpha*gamma*4*deltaR*np.exp(-deltaT/(sc.k*ts))*np.sinh(zeta)*k0(zeta)

    #Get the difference from the total Q and
    model = 1.0/(invQtls + invQmbd + 1.0/q0)

    #Weight the residual if eps is supplied
    if data is not None:
        if eps is not None:
            residual = (model-data)/eps
        else:
            residual = (model-data)

        return residual
    else:
        return model



def f0_tlsAndMBT(params, temps, powers, data = None, eps = None, **kwargs):
    """A model of frequency shift vs temperature and power, weighted by uncertainties.

    Parameters
    ----------
    params : ``lmfit.Parameters`` object
        Parameters must include ``['Fd', 'df', 'fRef', 'alpha', 'delta0']``.

    temps : ``numpy.Array``
        Array of temperature values to evaluate model at. May be 2D.

    powers : ``numpy.Array``
        Array of power values to evaluate model at. May be 2D.

    data : ``numpy.Array``
        Data values to compare to model. May also be ``None``, in which case
        function returns model.

    eps : ``numpy.Array``
        Uncertianties with which to weight residual. May also be ``None``, in
        which case residual is unwieghted.

    Returns
    -------

    residual : ``numpy.Array``
        The weighted or unweighted vector of residuals if ``data`` is passed.
        Otherwise, it returns the model.

    Note
    ----
    The following constraint must be satisfied::

        all(numpy.shape(x) == numpy.shape(data) for x in [temps, powers, eps])

    It is almost certain that this model does NOT apply to your device as the
    assumptions it makes are highly constraining and ignore several material
    parameters. It is included here more as an example for how to write a model
    than anything else, and it does at least qualitatively describe the behavior
    of most superconducting resonators.

    This model is taken from J. Gao's Caltech dissertation (2008) and the below
    equations are from that work.

    (2.54) gives for MBD (f(T)-f(0))/f(0) = -alpha*0.5*(X(T)-X(0))/X(0)

    (5.71) gives for TLS "" = Fd/pi * (usual TLS expression from Phillips)

    (X(T)-X(0))/X(0) calculated from (2.80), (2.89), and (2.90), using the ``deltaBCS``
    function in this module for returning gap as a function of temperature.

    """
    #Unpack parameter values from params
    Fd = params['Fd'].value
    f0 = params['f0'].value
    alpha = params['alpha'].value
    delta0 = params['delta0'].value*sc.e

    #Set temperature units
    units = kwargs.pop('units', 'mK')
    assert units in ['mK', 'K'], "Units must be 'mK' or 'K'."

    if units == 'mK':
        ts = temps*0.001

    #Calculate tc from BCS relation
    tc = delta0/(1.76*sc.k)

    #Get the reduced energy gap
    deltaR = deltaBCS(ts/tc)

    #And the energy gap at T
    deltaT = delta0*deltaR

    #Pack all these together for convenience
    zeta = sc.h*f0/(2*sc.k*ts)

    #Assuming thick film local limit
    #Other good options are 1 or 1/3
    gamma = kwargs.pop('gamma', 0.5)

    #TLS contribution
    dfTLS = Fd/sc.pi*(np.real(digamma(0.5+zeta/(1j*sc.pi)))-np.log(zeta/sc.pi))

    #MBD contribution
    dfMBD = alpha*gamma*0.5*(deltaR*(1-
                        2*np.exp(-deltaT/(sc.k*ts))
                        *np.exp(-zeta)
                        *i0(zeta))-1)


    #Calculate model from parameters
    model = f0+f0*(dfTLS + dfMBD)


    #Weight the residual if eps is supplied
    if data is not None:
        if eps is not None:
            residual = (model-data)/eps
        else:
            residual = (model-data)

        return residual
    else:
        return model

@np.vectorize
def deltaBCS(temp):
    r"""Return the reduced BCS gap deltar = delta(T)/delta(T=0).

    Parameters
    ----------
    temp : float
        reduced temperature t=(T/Tc) where Tc is critical temperature.

    Returns
    -------
    deltar : float
        Superconducting gap, delta, normalized by delta(T=0)

    Note
    ----
    Function interpolates data from Muhlschlegel (1959). For temperature below
    1.8 K, the following functional form is used::

        gap = np.exp(-np.sqrt(3.562*temp)*np.exp(-1.764/temp))

    For temperatures below 50 mK, it returns 1.

    """

    #These data points run from t=0.18 to t=1
    #in steps of 0.02 from Muhlschlegel (1959)
    delta_calc = [1.0, 0.9999, 0.9997, 0.9994, 0.9989,
           0.9982, 0.9971, 0.9957, 0.9938, 0.9915,
           0.9885, 0.985,  0.9809, 0.976,  0.9704,
           0.9641, 0.9569, 0.9488, 0.9399, 0.9299,
           0.919,  0.907,  0.8939, 0.8796, 0.864,
           0.8471, 0.8288, 0.8089, 0.7874, 0.764,
           0.7386, 0.711,  0.681,  0.648,  0.6117,
           0.5715, 0.5263, 0.4749, 0.4148, 0.3416,
           0.2436, 0.0]

    t_calc = np.linspace(0.18, 1, len(delta_calc))

    if 1 >= temp >= 0.3:
        #interpolate data from table
        gap = float(si.interp1d(t_calc, delta_calc, kind='cubic')(temp))
    elif temp > 1 or temp < 0:
        gap = 0.0
    elif temp < 0.05:
        gap = 1.0
    else:
        gap = np.exp(-np.sqrt(3.562*temp)*np.exp(-1.764/temp))

    return gap



from setuptools import setup, find_packages
import os

version_file = open(os.path.join('.', 'VERSION.txt'))
version_number = version_file.read().strip()
version_file.close()

setup(
    name = 'scraps',
    description = 'SuperConducting Resonator Analysis and Plotting Software.',
    version = version_number,
    author = 'Faustin Carter',
    author_email = 'faustin.carter@gmail.com',
    license = 'MIT',
    url = 'http://github.com/FaustinCarter/scraps',
    packages = ['scraps', 'scraps.fitsS21', 'scraps.fitsSweep'],
    long_description = open('README.rst').read(),
    install_requires = [
        'numpy>=1.5',
        'matplotlib>=2.0',
        'scipy>=0.14',
        'lmfit>=0.9.5',
        'emcee>=2.2.1',
        'pandas>=0.18'
    ],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Visualization'
    ]

)
